#!/usr/bin/perl
#
#  This script "converts" actual Cassini attitude and rate data
#  file in query2plot output ECSV format to a table acceptable for
#  MSOPCK.
#
#  The script can be run in two ways:
#
#     1) as a UNIX filter that reads input from STDIN and writes output
#        to STDOUT. In this case the input file contents should be
#        re-directed into the script from STDIN and the STDOUT output
#        from the script should be re-directed to the output file, like
#        this:
#
#           % cas_ecsv2msopck.pl < [ecsv_att_file] > [msopck_inp_file]
#
#     2) by taking input from a file specified as the first argument on
#        the command line and writing output to a file specified as
#        the second argument on the command line:
#
#           % cas_ecsv2msopck.pl [ecsv_att_file] [msopck_inp_file]
#
#  A predict attitude file consists of time ordered attitude/rate
#  data points in the following format:
#
#     ...(SFDU and other header information)
#     2000-033T20:37:32.313,A-1001,ATEquat1,15730,0.480056,1328215556.096,I,,,,,2000-033T21:02:10.582
#     2000-033T20:37:32.313,A-1002,ATEquat2,14878,0.454054,1328215556.096,I,,,,,2000-033T21:02:10.582
#     2000-033T20:37:32.313,A-1003,ATEquat3,22424,0.684347,1328215556.096,I,,,,,2000-033T21:02:10.582
#     2000-033T20:37:32.313,A-1004,ATEquat4,10102,0.308298,1328215556.096,I,,,,,2000-033T21:02:10.582
#     2000-033T20:37:32.313,A-1005,X_rate,5,2.5e-05,1328215556.096,I,,,,,2000-033T21:02:10.582
#     2000-033T20:37:32.313,A-1006,Y_rate,5,2.5e-05,1328215556.096,I,,,,,2000-033T21:02:10.582
#     2000-033T20:37:32.313,A-1007,Z_rate,-1,-5e-06,1328215556.096,I,,,,,2000-033T21:02:10.582
#     2000-033T20:37:36.313,A-1001,ATEquat1,15729,0.480026,1328215560.096,I,,,,,2000-033T21:02:10.685
#     2000-033T20:37:36.313,A-1002,ATEquat2,14878,0.454054,1328215560.096,I,,,,,2000-033T21:02:10.685
#     2000-033T20:37:36.313,A-1003,ATEquat3,22425,0.684378,1328215560.096,I,,,,,2000-033T21:02:10.685
#     2000-033T20:37:36.313,A-1004,ATEquat4,10102,0.308298,1328215560.096,I,,,,,2000-033T21:02:10.685
#     2000-033T20:37:36.313,A-1005,X_rate,-6,-3e-05,1328215560.096,I,,,,,2000-033T21:02:10.685
#     2000-033T20:37:36.313,A-1006,Y_rate,1,5e-06,1328215560.096,I,,,,,2000-033T21:02:10.685
#     2000-033T20:37:36.313,A-1007,Z_rate,2,1e-05,1328215560.096,I,,,,,2000-033T21:02:10.685
#     ...
#
#  where Channels A-1001 thru A-1004 are quaternions, channels A-1005
#  thru A-1007 are angular rates.
#
#  The output table for such input will look like:
#
#     1328215556.096      0.480056     0.454054     0.684347     0.308298      2.5e-05      2.5e-05       -5e-06
#     1328215560.096      0.480026     0.454054     0.684378     0.308298       -3e-05        5e-06        1e-05
#     ...
#
#  Any input lines that are not quaternion or angular rate data
#  lines (including SFDU header lines) and any "incomplete"
#  attitude record blocks (with one or more quaternion or angular
#  rate components missing) are filtered out and written to the
#  "cas_ecsv2msopck.bad_lines" file.
#
#  By Boris Semenov/NAIF, x4-8136
#
#  Version 3.0 -- March 29, 2002
#
#     Scripts interface was extended to allow running it two ways:
#
#        -- as a UNIX filter, i.e. reads input from STDIN, write output
#           to STDOUT (this is the way versions 1.0 and 2.0 worked);
#
#        -- by taking input from a file specified as the first argument on
#           the command line and writing output to a file specified as
#           the seconds argument on the command line.
#
#  Version 2.0 -- January 8, 2001
#
#     Fixed value filter line to allow for decimals. Added re-direction of
#     "non-data" input lines and "incomplete" data sets to a "bad lines" file.
#
#  Version 1.0 -- February 8, 2000
#

#
#  Script name and "bad lines" file name.
#
$scriptname = "cas_ecsv2msopck.pl";
$badlinfile = "cas_ecsv2msopck.bad_lines";

#
#  Process command line. It must have zero or two arguments. If it has zero
#  arguments, run in filter mode (input from STDIN, output to STDOUT.) If
#  it has two arguments, the first one is the name of the input file and
#  the second one is the name of the output file.
#
if      ( @ARGV == 2 ) {
   $runmode = "files";
   $infile  = shift @ARGV;
   $outfile = shift @ARGV;
} elsif ( @ARGV == 0 ) {
   $runmode = "filter";
} else {
   die "\nERROR: $scriptname: wrong number of command line arguments.\n\n";
}

#
#  Assign input and output stream units based on the mode.
#
if      ( $runmode eq "files"  ) {

   #
   #  Try to open files.
   #
   open ( INFILE, "<$infile" )
      || die "\nERROR: $scriptname: cannot open $infile for read access.\n";
   open ( OUTFILE, ">$outfile" )
      || die "\nERROR: $scriptname: cannot open $outfile for write access.\n";

   #
   #  Assign input and output stream "units" to opened files.
   #
   $INPUT = INFILE;
   $OUTPUT = OUTFILE;

} elsif ( $runmode eq "filter" ) {

   #
   #  Assign input and output stream "units" to STDIN/STDOUT.
   #
   $INPUT = STDIN;
   $OUTPUT = STDOUT;

} else {
   die "\nERROR: $scriptname: just hit a bug ...\n\n";
}

#
#  Initialize record values buffer.
#
$value{"A-1001"} = "";
$value{"A-1002"} = "";
$value{"A-1003"} = "";
$value{"A-1004"} = "";
$value{"A-1005"} = "";
$value{"A-1006"} = "";
$value{"A-1007"} = "";

#
# Open file that will capture "bad" records.
#
open ( BADLINES, ">$badlinfile" )
   || die "\nERROR: $scriptname: cannot open $badlinfile file for bad lines.\n";

#
#  We will count input lines,
#
$count = 0;

#
#  save time tag from the last processed line,
#
$lasttime = 0;

#
#  keep a flag indicating that we we just wrote a complete record,
#
$wroteline = "TRUE";

#
#  and take lines from input stream unit.
#
while( $line = <$INPUT> ) {

   #
   #  Split current line.
   #
   $count++;
   chop( $line );
   @tokens = split( ',', $line );

   #
   #  See if second, fifth and sixth tokens contain channel ID, value
   #  and SCLK. (Note that regular expression for value -- $tokens[4] -- is
   #  pretty loose compared to the other two .. but it will make sure that
   #  only numbers will get though :-)
   #
   if ( $tokens[1] =~ /^A-100[1-7]$/ &&
        $tokens[4] =~ /(^(-|\+)?\d{1,1}(\.)?(\d{1,16})?$|^(-|\+)?\d{1,1}(\.\d{1,16})?(e|E)(-|\+)?\d{1,6}$)/ &&
        $tokens[5] =~ /^\d{10,10}\.\d{3,3}$/  ) {

      #
      #  If we moved onto the next "block" but didn't write the
      #  the previous one, we probably had incomplete record.
      #
      if ( $lasttime ne $tokens[5] && $wroteline ne "TRUE" ) {
         print BADLINES "INCOMPLETE_ATT_REC(".$lnums{"A-1001"}."): ".$lines{"A-1001"}."\n";
         print BADLINES "INCOMPLETE_ATT_REC(".$lnums{"A-1002"}."): ".$lines{"A-1002"}."\n";
         print BADLINES "INCOMPLETE_ATT_REC(".$lnums{"A-1003"}."): ".$lines{"A-1003"}."\n";
         print BADLINES "INCOMPLETE_ATT_REC(".$lnums{"A-1004"}."): ".$lines{"A-1004"}."\n";
         print BADLINES "INCOMPLETE_ATT_REC(".$lnums{"A-1005"}."): ".$lines{"A-1005"}."\n";
         print BADLINES "INCOMPLETE_ATT_REC(".$lnums{"A-1006"}."): ".$lines{"A-1006"}."\n";
         print BADLINES "INCOMPLETE_ATT_REC(".$lnums{"A-1007"}."): ".$lines{"A-1007"}."\n";
      }

      #
      #  This is a data line -- save tokens in the associative arrays.
      #
      $time{$tokens[1]}  = $tokens[5];
      $value{$tokens[1]} = $tokens[4];
      $lines{$tokens[1]} = $line;
      $lnums{$tokens[1]} = $count;

      #
      #  See if we collected a complete record -- all channels are
      #  present and all time tags are the same.
      #
      if ( (   $value{"A-1001"} ne "" &&
               $value{"A-1002"} ne "" &&
               $value{"A-1003"} ne "" &&
               $value{"A-1004"} ne "" &&
               $value{"A-1005"} ne "" &&
               $value{"A-1006"} ne "" &&
               $value{"A-1007"} ne ""                 ) &&

            (  $time {"A-1001"} eq $time{"A-1002"} &&
               $time {"A-1001"} eq $time{"A-1003"} &&
               $time {"A-1001"} eq $time{"A-1004"} &&
               $time {"A-1001"} eq $time{"A-1005"} &&
               $time {"A-1001"} eq $time{"A-1006"} &&
               $time {"A-1001"} eq $time{"A-1007"}    )     ) {

         #
         #  We have got a complete record (in which each item is
         #  with the same time tag :-) So, print it out.
         #
         printf $OUTPUT "%-15s %20s %20s %20s %20s %20s %20s %20s\n",
               $time {"A-1001"},
               $value{"A-1001"},
               $value{"A-1002"},
               $value{"A-1003"},
               $value{"A-1004"},
               $value{"A-1005"},
               $value{"A-1006"},
               $value{"A-1007"};

         #
         #  Reset the buffers.
         #
         $value{"A-1001"} = "";
         $value{"A-1002"} = "";
         $value{"A-1003"} = "";
         $value{"A-1004"} = "";
         $value{"A-1005"} = "";
         $value{"A-1006"} = "";
         $value{"A-1007"} = "";

         $lines{"A-1001"} = "";
         $lines{"A-1002"} = "";
         $lines{"A-1003"} = "";
         $lines{"A-1004"} = "";
         $lines{"A-1005"} = "";
         $lines{"A-1006"} = "";
         $lines{"A-1007"} = "";

         $lnums{"A-1001"} = "";
         $lnums{"A-1002"} = "";
         $lnums{"A-1003"} = "";
         $lnums{"A-1004"} = "";
         $lnums{"A-1005"} = "";
         $lnums{"A-1006"} = "";
         $lnums{"A-1007"} = "";

         #
         #  Reset flag to indicate that we wrote an output line.
         #
         $wroteline = "TRUE";

      } else {

         #
         #  Reset flag to indicate that we didn't write the line.
         #
         $wroteline = "FALSE";

      }

      #
      #  Memorize current time.
      #
      $lasttime = $tokens[5];

   } else {

      #
      #  This line didn't get parsed as a data line. Report this.
      #
      print BADLINES "NOT_AN_ATT_LINE($count): $line\n";

   }

}

#
#  Close "bad lines" file.
#
close ( BADLINES )
   || die "\nERROR: $scriptname: cannot close $badlinfile file.\n";

#
#  Close input and output files if we run in "files" mode.
#
if ( $runmode eq "files" ) {
   close ( INFILE  )
      || die "\nERROR: $scriptname: cannot close $infile.\n";
   close ( OUTFILE )
      || die "\nERROR: $scriptname: cannot close $outfile.\n";
}

#
#  All done.
#
exit;
